/************************************************************************/
/*  msomerge -	merge multiple miniseed opaque data streams into single	*/
/*		continuous output data stream.				*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: msomerge.c,v 1.3 2000/10/26 13:39:25 doug Exp $ ";
#endif

#include <stdio.h>
#include "version.h"

#include "qlib2.h"

#include "params.h"

char *syntax[] = {
"%s version " VERSION,
"%s  [-o output_file | -n] [-T] [-f date] [-t date | -s interval] [-e]",
"    [-u] [-P duration] [-pn | -pd]",
"    [-g tol] [-G tol] [-V version] [-x blockette_list|all]",
"    [-S station] [-C channel] [-N network] [-L location]",
"    [-D j|m] [-v] [-b blksize] [-i time] [-i d0] [-i no_d0] [-c] [-H] [-d n]",
"    [-F file_list_file] [-O n] [-I input_data_fmt] [-h]",
"    [input_file_1 ... input_file_n]",
"    where:",
"	-o outfile  Name of output file (or template filename with -P option).",
"	-a	    Append output to file instead of overwriting file.",
"	-n	    No data output.  Equivalent to -o /dev/null.",
"	-T	    Exact trim -- trim data to closest specified time.",
"		    Default is to trim to nearest inclusive block boundary.",
"	-f date	    From date - ignore data before this date.",
"	-t date	    To date - ignore data from this date on.",
"		    Date can be in formats:",
"			yyyy/mm/dd/hh:mm:ss.ffff",
"			yyyy/mm/dd.hh:mm:ss.ffff",
"			yyyy/mm/dd,hh:mm:ss.ffff",
"			yyyy.ddd,hh:mm:ss.ffff",
"			yyyy,ddd,hh:mm:ss.ffff",
"		    You may leave off any trailing 0 components of the time.",
"	-s interval span interval.  Alternate way of specifying end time.",
"		    Interval can be an integer followed immediately by",
"		    s, m, h, or d for seconds, minutes, hours, or days.",
"	-e	    Output only data blocks that have event flags turned on.",
"	-u	    Unique data flag.  Assume ALL input data is unique,",
"		    and suppress the removal of apparent duplicate data caused",
"		    by apparent data overlaps between data records.",
"	-P duration Partition output into distinct files of specified duration.",
"		    Duration may be 'NU' where N = decimal number, and",
"		    U is one of the following characters: y = year,",
"		    m = months, d = days, H = hours, M = minutes, S = seconds.",
"	-pn | -pd   Create name for partitioned output file based on nominal",
"		    partition time or on first data time.",
"		    Default is based on nominal parition time.",
"	-b blksize  Specify output blksize in bytes.",
"		    Must be between 256 and 8192 inclusive, and blksize = 2^N.",
"	-g tol	    Tolerance between blockette numbers for determining whether",
"		    blockette numbering has been reset within a single stream.",
"		    Default is " STRING(DEFAULT_BLOCK_TOL) ".",
/*
"	-G TOL	    Explicit tolerance for time gaps in ticks (1/10 msec).",
"		    Default is " STRING(TTOL_MULT) " of data sample interval.",
*/
"	-V version  Specify SEED version number for data header and blockettes.",
"		    Default version is 2.2",
"	-x list|all Strip the specified blockettes or all blockettes.",
"		    List can be comma-delimted and include a span, such as:",
"		    200-500,1001 or the string 'all' for all blockettes.",
"	-S station  Set the 5 char station name to the specified station name.",
"	-C channel  Set the 3 char channel name to the specified channel name.",
"	-N network  Set the 2 char network code (eg BK).",
"	-L location Set the 2 char location code (normally blank).",
"	-D j | m    Display dates in julian or month and day format.",
"	-i time	    Ignore time span info in header.",
"	-i d0	    Ignore (silently fix) d0 discrepencies between blocks.",
"	-i no_d0    Ignore and do NOT fix d0 discrepencies between blocks",
"		    (unless data is reblocked).",
"	-v	    Verify internal consistency of all data frames and blocks.",
"	-c	    Write a content index line of station, channel, network,",
"		    location, and blksize of the first file to stdout.",
"		    Blank network and location code will be displayed as '--'.",
"		    Do not process files.",
"	-H	    Write content index info and file header timespan info",
"		    to stdout for each file.  Timespan will be displayed only.",
"		    for SEED telemetry files.  Do not process files.",
"	-h	    Help - prints syntax message.",
"	-d n	    Debug output (OR of any of the following values:)",
"			1 = time slews",
"			2 = time info",
"			4 = block info",
"			8 = block headers",
"			16 = flags",
"			32 = input stream",
"			64 = output stream",
"			1024 = output ascii",
"	-F file_list_file",
"		    File containing a list of data files, one file per line.",
"		    If this option is used, no input_file names are read from",
"		    the command line.",
"	-O output_data_wordorder",
"		    Specifies the output word order for the miniSEED data.",
"		    Default is the wordorder of the machine",
"		    Possible values are: 0 (LITTLE_ENDIAN) and 1 (BIG_ENDIAN),",
"		    SPARC and Motorola are BIG_ENDIAN, Intel is LITTLE_ENDIAN.",
"	-I input_data_format",
"		    Specifies the input data format of non-MiniSEED records.",
"		    Valid formats are STEIM1, STEIM2, INT_16, INT_32, INT_24.",
"		    Default format for non-MiniSEED records is STEIM1.",
"	input_file(s)",
"		    Input file(s) containing quanterra data.",
"		    If no input file is specified, input is read from stdin.",
NULL };

/************************************************************************/

#include <string.h>
#include <math.h>

#include "st_info.h"
#include "procs.h"
#define	EXTERN
#define	DEFINE
#include "externals.h"


/************************************************************************/
/*  main:   main program.						*/
/************************************************************************/
main
   (int		argc,
    char	**argv)
{
    ST_INFO	*st_p;		/* ptr to input stream structure.	*/
    ST_INFO	*out_p;		/* ptr to output stream structure.	*/
    ST_INFO	*prev_stream;	/* ptr to prev input stream structure.	*/
    int		i;
    int		bias, start_bias, run_bias;
    INT_TIME	*stime;
    INT_TIME	*etime;
    int		second, ticks;
    int		new_stream;
    struct {
	char	station[8];
	char	channel[8];
	char	network[4];
	char	location[4];
	int	blksize;
    }  index;

    /*	Initialization.							*/
    outfile = NULL;		/* empty initial output directory.	*/
    fpin = stdin;		/* assume input to come from stdin.	*/
    info = stdout;
    date_fmt = JULIAN_FMT;
    run_bias = NO_BIAS;
    sec_per_year(1970);		/* force leapsecond initialization.	*/

    debug_option = 0;

    if (parse_cmdline (&argc, &argv)) {
	fprintf (info, "Error: parsing command line\n");
	print_syntax(cmdname,syntax,info);
	exit(1);
    }

    if (debug(DEBUG_TIME)) {
	if (start_flag)
	    fprintf (info, "interval start time = %s\n", time_to_str(start_time,date_fmt));
	if (end_flag)
	    fprintf (info, "interval stop time = %s\n", time_to_str(end_time,date_fmt));
    }

    /*	Determine the block size and format of each input stream.	*/
    for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	int status;
	status = get_stream_info (st_p);
	if (status == EOF) {
	    /*  Save stream and channel info if not null.		*/
	    /*	All files may be empty of data, but a DRM file will	*/
	    /*	have the station and channel in the DRM header.		*/
	    if ((int)strlen(st_p->station) > 0) {
		strcpy (index.station, st_p->station);
		strcpy (index.channel, st_p->channel);
		strcpy (index.network, st_p->network);
		strcpy (index.location, st_p->location);
		index.blksize = st_p->blksize;
	    }
	    st_p = close_stream(st_p);
	}
    }

    /*  If index is desired, return info from first available file.	*/
    if (make_index) {
	if ((st_p = st_head_in.next) != NULL) {
	    fprintf (stdout, "%-5.5s %-3.3s %-2.2s %-2.2s %4d\n", 
		     st_p->station, st_p->channel, 
		     (st_p->network[0]) ? st_p->network : "--",
		     (st_p->location[0]) ? st_p->location : "--",
		     st_p->blksize);
	    exit(0);
	}
	else if ((int)strlen(index.station) > 0) {
	    fprintf (stdout, "%-5.5s %-3.3s %-2.2s %-2.2s %4d\n", 
		     index.station, index.channel,
		     (index.network[0]) ? index.network : "--",
		     (index.location[0]) ? index.location : "--",
		     index.blksize);
	    exit(0);
	}
	else {
	    fprintf (stderr, "Error: no index info for file\n");
	    exit(1);	/* No info for index is an ERROR.		*/
	}
    }

    /* If vol_file_hdr is desired, return info from all files.		*/
    if (print_file_hdr) {
	char station_id[6], location_id[3], channel_id[4], network_id[3];
	char reclen_str[3];
	char str_date[40];
	int blksize;
	char *p, *q;
	INT_TIME it[2], *pi;
	int i;
	for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	    if (st_p->vfh) {
		memset(station_id, 0, 6);
		memset(channel_id, 0, 4);
		memset(network_id, 0, 3);
		memset(location_id, 0, 3);
		memset(reclen_str, 0, 1);
		strncpy(reclen_str,st_p->vfh+19,2);
		blksize = (int)pow(2.0,atoi(reclen_str));
		trim(strncpy(station_id,st_p->vfh+21,5));
		trim(strncpy(location_id,st_p->vfh+26,2));
		trim(strncpy(channel_id,st_p->vfh+28,3));
		p = st_p->vfh+31;
		/* Decode begin and end dates. */
		for (i=0; i<2; i++) {
		    q = strchr(p,'~');
		    strncpy(str_date,p,q-p);
		    str_date[q-p] = '\0';
		    if (pi = parse_date(str_date)) {
			it[i] = *pi;
		    }
		    else {
			fprintf (info, "Error: Invalid volume hdr\n");
			exit(1);
		    }
		    p=q+1;
		}
		p=strchr(p,'~')+1;
		p=strchr(p,'~')+1;
		trim(strncpy(network_id, p, 2));
		fprintf (stdout, "%s.%s.%s.%s %d %s to ",
			 station_id, 
			 network_id,
			 channel_id, 
			 location_id,
			 blksize, time_to_str(it[0],date_fmt+1));
/*::		
		fprintf (stdout, "%-5.5s %-3.3s %-2.2s %s to ",
			 station_id, channel_id, network_id,
			 time_to_str(it[0],date_fmt+1));
::*/
		fprintf (stdout, "%s\n", time_to_str(it[1],date_fmt+1));
	    }
	    else {
		fprintf (stdout, "%s.%s.%s.%s %d \n", 
			 st_p->station, st_p->network, st_p->channel,
			 st_p->location, st_p->blksize);
/*::
		fprintf (stdout, "%-5.5s %-3.3s %-2.2s \n", 
			 st_p->station, st_p->channel, st_p->network);
::*/
	    }
	    st_p = close_stream (st_p);
	    continue;
	}
	exit(0);
    }

    /*	Get the first block for each stream.				*/
    for (st_p = st_head_in.next; st_p != NULL; st_p = st_p->next) {
	get_block_info (st_p);
    }

    /*  Start with the file that has the earliest record time within	*/
    /*  the specified time window (or earliest time if no start time).	*/
    /*	After the first record, look for the next consecutive		*/
    /*  blockette 2000 record number.					*/
    /*  Stay with the current file if it contains the desired blockette.*/

    start_bias = (start_flag) ? START_BIAS : run_bias;
    bias = start_bias;

    /*	Initialize search parameters.					*/
    st_p = NULL;
    out_p = st_head_out.next;
    stime = (start_flag) ? &start_time : NULL;
    etime = (end_flag) ? &end_time : NULL;

    /*	Loop reading input and creating output.				*/
    /*	Now find the "best" stream to use.				*/
    /*	By definition, all streams have current block info.		*/
    prev_stream = st_p;
    while ( (st_p = find_best_stream (st_p, out_p, stime, etime, bias)) != NULL ) {
	/*	Process the data in the current block.			*/
	new_stream = (prev_stream != st_p);
	if (new_stream) {
	    if (debug(DEBUG_STREAM)) fprintf (info, "time %s using stream %s\n",
		time_to_str(st_p->cur_hdr->begtime,date_fmt), st_p->filename);
	    prev_stream = st_p;
	}
	prev_stream = st_p;
	process_block (st_p, new_stream);
	out_p = st_head_out.next;
	stime = NULL;
	bias = run_bias;
    }
    flush_output(out_p);
    print_stream_summary(out_p);
    return(0);
}

/************************************************************************/
/*  print_stream_summary:						*/
/*	Print stream summary info.					*/
/************************************************************************/
void print_stream_summary
   (ST_INFO	*out_p)		/* ptr to stream structure.		*/
{
    int		seconds, usecs;
    INT_TIME	endtime;

    if (out_p->sum_hdr != NULL) {
	/*	Output summary information for the output stream.		*/
	time_interval2 (1, out_p->sum_hdr->sample_rate, out_p->sum_hdr->sample_rate_mult,
			&seconds, &usecs);
	fflush (out_p->iob->fp);
	endtime = add_time (out_p->sum_hdr->endtime, seconds, usecs);
	fprintf (info, "%s %s: data (%s to ",
		 out_p->sum_hdr->station_id, out_p->sum_hdr->channel_id, 
		 time_to_str(out_p->sum_hdr->begtime,date_fmt));
	fprintf (info, "%s) : ", time_to_str(endtime,date_fmt));
	fprintf (info, "(%d to %d) : ", 
		 out_p->sum_hdr->xm1, out_p->sum_hdr->xm2);
	fprintf (info, "%d blockettes\n", out_p->sum_hdr->num_samples);

    }
}
